// Horizontal-Generator
// fr Philips BRE-Modul 
// Frank Brennecke, 05.04.2013
// Version 1.2
// Fr ATMega88 mit 20-MHz-Quarz
// Fuses: Extended F9, High: DE, Low: F7

// horizontale Ausgangsfrequenzen an C0: 38,4 KHz (VGA800), 31,4 kHz (VGA640), 15,625 kHz (FBAS)
// Zustzliche Ausgangsfrequenzen bei internem Modus: fx1, fx2, fx4, fx8 bei Taktverhltnis 1:1

// vertikale AUsgangsfrequenzen an C1: 60 Hz (VGA800), 60 Hz (VGA640), 50 Hz (FBAS)
// Zustzliche Ausgangsfrequenzen bei internem Modus: fx1, fx2, fx4, fx8 an C2-C5
// Der Vertikalgenerator bentigt aufgrund der geringen internen Frequenzen keinen eigenen Quarz-Oszillator, 
//   sondern kann mit der internen Frequenzeinheit betrieben werden. Beim Betrieb als interner Generator 
//   wird er automatisch mit dem Horizontalgenerator synchronisiert. Er besitzt daher keinen eigenen internen Modus.


#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/io.h>
#include <avr/wdt.h>

// Modi fr externe Sync-Signale
#define VGA800 		0x1C	// SVGA 800x600 selected
#define VGA640 		0x1A 	// VGA 640x480 selected
#define FBAS 		0x16 	// FBAS selected
// Interner Generator - diese Betriebsart macht aus der Schaltung entweder ein Video-Modul (analog zum EE2023) oder ein VGA-Modul 
#define INTGEN_VGA	0x0A	// Internal VGA Signal Generator active
#define INTGEN_FBAS	0x06	// Internal FBAS Signal Generator active

#define VGA800_IN	0x01	// SVGA 800x600 erkannt, setzt Ausgang D0
#define VGA640_IN	0x02 	// VGA 640x480 erkannt, Eingang D1 (per Schalter)
// FBAS-Signal wird erkannt, wenn D1 "0" ist

#define INTERN_GEN		0x40	// Interner Generator aktiviert, Eingang D6 (wird vom Vertrikalgenerator erkannt)

#define INT_CONTROL			EICRA
#define INT0_FALLING_EDGE	0x02 	//Bit 1 gesetzt 
#define INT0_RISING_EDGE	0x03	//Bit 0 und Bit 1 gesetzt
#define INT1_FALLING_EDGE	0x08	//Bit 3 gesetzt
#define INT1_RISING_EDGE	0x0C	//Bit 2 und Bit 3 gesetzt

#define INT_ENABLE			EIMSK
#define INT0_ENABLE			0x01	//Bit 0 gesetzt
#define INT1_ENABLE			0x02	//Bit 1 gesetzt
#define INT_DISABLE			0x00	//kein Bit gesetzt


// Krzel fr schnelle Wartezeiten (1 Takt)
#define NOP asm("nop")
#define NOP10 asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop");  asm("nop"); asm("nop");
#define NOP20 asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop");
#define NOP21 asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop");
#define NOP8 asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop"); asm("nop");  asm("nop"); asm("nop"); 
#define NOP4 asm("nop"); asm("nop"); asm("nop"); asm("nop"); 


//Makro for video handle
//HSYNC
#define hsync_on  PORTC = 0x01;   // C0 (HS)
#define hsync_off  PORTC = 0x00;   // all off (HS)
 
//horizontal Outputs
#define h_2 PORTC = 0x04; // C2 (Q1)
#define h_3 PORTC = 0x08; // C3 (Q2)
#define h_4 PORTC = 0x0C; // C2 and C3 (Q1 and Q2)
#define h_5 PORTC = 0x10; // C4 (Q3)
#define h_6 PORTC = 0x14; // C4 and C2 (Q3 and Q1)
#define h_7 PORTC = 0x18; // C4 and C3 (Q3 and Q2)
#define h_8 PORTC = 0x1C; // C4 and C3 and CD2 (Q3 and Q2 and Q1)
#define h_9 PORTC = 0x20; // C5 (Q4)
#define h_10 PORTC = 0x24; // C5 and C2 (Q4 and Q1)
#define h_11 PORTC = 0x28; // C5 and C3 (Q4 and Q2)
#define h_12 PORTC = 0x2C; // C5 and C2 and C3 (Q4 and Q1 and Q2)
#define h_13 PORTC = 0x30; // C5 and C4 (Q4 and Q3)
#define h_14 PORTC = 0x34; // C5 and C4 and C2 (Q4 and Q3 and Q1)
#define h_15 PORTC = 0x38; // C5 and C4 and C3 (Q4 and Q3 and Q2)
#define h_16 PORTC = 0x3C; // C5 and C4 and C3 and C2 (Q4 and Q3 and Q2 and Q1)
#define h_off PORTC = 0x00; // all off

//vertical Outputs
#define v_2 PORTB = 0x02; // B1 (Q5)
#define v_3 PORTB = 0x04; // B2 (Q6)
#define v_4 PORTB = 0x06; // B1 and B2 (Q5 and Q6)
#define v_5 PORTB = 0x08; // B3 (Q7)
#define v_6 PORTB = 0x0A; // B3 and B1 (Q7 and Q5)
#define v_7 PORTB = 0x0C; // B3 and B2 (Q7 and Q6)
#define v_8 PORTB = 0x0E; // B3 and B1 and B2 (Q7 and Q6 and Q5)
#define v_9 PORTB = 0x10; // B4 (Q8)
#define v_10 PORTB = 0x12; // B4 and B1 (Q8 and Q5)
#define v_11 PORTB = 0x14; // B4 and B2 (Q8 and Q6)
#define v_12 PORTB = 0x16; // B4 and B1 and B2 (Q8 and Q5 and Q6)
#define v_13 PORTB = 0x18; // B4 and B3 (Q8 and Q7)
#define v_14 PORTB = 0x1A; // B4 and B3 and B1 (Q8 and Q7 and Q5)
#define v_15 PORTB = 0x1C; // B4 and B3 and B2 (Q8 and Q7 and Q6)
#define v_16 PORTB = 0x1F; // B4 and B3 and B2 and B1 (Q8 and Q7 and Q6 and Q5) and VSync
#define v_17 PORTB = 0x01; // VSync still active
#define v_off PORTB = 0x00; // all off

#define VGA640_lines_internal  	525 // VGA-Signal hat 525 Zeilen
#define FBAS_lines_internal	 	313 // FBAS-Signal hat 2x312,5 Zeilen
#define VGA640_lines			526 // VGA640-Signal wird um eine Zeile verlngert, damit der V-Sync greifen kann
#define VGA800_lines			626 // VGA800-Signal wird um eine Zeile verlngert, damit der V-Sync greifen kann
#define FBAS_lines				314 // FBAS-Halbbild wird um eine Zeile verlngert, damit der V-Sync greifen kann

volatile unsigned int sync_mode;   // Welcher Sync-Modus ist eingestellt?
volatile unsigned int max_delay_time; // Maximale Verschiebung der Sync-Position (Bildlage)
volatile unsigned int porch_delay;  // Tatschliche Verschiebung der Sync-Position (Bildlage)
volatile unsigned int linecount; // Anzahl der durchlaufenen Zeilen, fr Generierung des internen VSync-Signals

volatile unsigned int vga_mode;



ISR(TIMER1_COMPA_vect)
{
	
	cli();
	
	
	
// Behandlungsroutine fr internen Generator
// Diese Routine wird aufgerufen, wenn ein Timer abluft, der genau eine Zeilenlnge hat (FBAS oder VGA)
// Es wird ein positiver Sync-Impuls passender Lnge erzeugt, der am Ausgang ausgegeben wird
// Es werden auerdem weitere Impulse hherer Frequenz erzeugt
	if (sync_mode == INTGEN_VGA) // Es soll ein VGA-Signal erzeugt werden
	{
		
	
		h_off;  
		NOP10;
		h_2;

		NOP10;
		NOP8;
	
		h_3;  
		NOP10;
		NOP8;

		h_4;
		NOP10;
		NOP8;
		
		h_5;  
		NOP10;
		NOP8;

		h_6;
		NOP10;
		NOP8;
	
		h_7;  
		NOP10;
		NOP8;
		
		h_8;
		NOP10;
		NOP8;
    
		h_9;  
		NOP10;
		NOP8;
	
		h_10;
		NOP10;
		NOP8;
	
		h_11;  
		NOP10;
		NOP4;
	
		h_12;
		NOP10;
		NOP8;
	
		h_13;  
		NOP10;
		NOP8;
	
		h_14;
		NOP10;
		NOP8;
	
		h_15;  
		NOP10;
		NOP8;
	
		h_16;
		NOP10;
		NOP8;
		
		h_off;
		NOP10;
		NOP20;
		NOP20;

// horizontal sync : 96 pixels = 127 Takte	
		hsync_on;
		NOP20;
		hsync_off;	
	
// Back Porch - use for vertical sync	
		if (++linecount == VGA640_lines_internal)
		{
			linecount = 0;   //Reset linecount on maximum
		} 

		switch (linecount)
		{
			case 22: 
				v_2;
				break;
			case 56:
				v_3;
				break;
			case 88: 
				v_4;
				break;
			case 121:
				v_5;
				break;
			case 154: 
				v_6;
				break;
			case 187:
				v_7;
				break;
			case 220: 
				v_8;
				break;
			case 253:
				v_9;
				break;
			case 285: 
				v_10;
				break;
			case 318: 
				v_11;
				break;
			case 351:
				v_12;
				break;
			case 384: 
				v_13;
				break;
			case 417:
				v_14;
				break;
			case 449: 
				v_15;
				break;
			case 481:
				v_16;
				break;
			case 515:
				v_17;
				break;
			case 524:
				v_off;
				break;	
		

		}
	
	}
	else  // es soll ein FBAS-Signal erzeugt werden
	{
		
		h_off;  
		NOP21;
		h_2;

		NOP21;
		NOP21;
		NOP8;
	
		h_3;  
		NOP21;
		NOP21;
		NOP8;

		h_4;
		NOP21;
		NOP21;
		NOP8;
		
		h_5;  
		NOP21;
		NOP21;
		NOP8;

		h_6;
		NOP21;
		NOP21;
		NOP8;
	
		h_7;  
		NOP21;
		NOP21;
		NOP8;
		
		h_8;
		NOP21;
		NOP21;
		NOP8;
    
		h_9;  
		NOP21;
		NOP21;
		NOP8;
	
		h_10;
		NOP21;
		NOP21;
		NOP8;
	
		h_11;  
		NOP21;
		NOP21;
		NOP4;
	
		h_12;
		NOP21;
		NOP21;
		NOP8;
	
		h_13;  
		NOP21;
		NOP21;
		NOP4;
	
		h_14;
		NOP21;
		NOP21;
		NOP8;
	
		h_15;  
		NOP21;
		NOP21;
		NOP4;
	
		h_16;
		NOP21;
		NOP21;
		NOP8;
	
		h_off;
		NOP20;
		NOP20;
		NOP20;
		NOP20;

// horizontal sync : 240 Takte = 12 s
		hsync_on;
		NOP8;
		NOP8;
		NOP20;
		NOP20;
		hsync_off;	
	
// Back Porch - use for vertical sync	
		if (++linecount == FBAS_lines_internal)
		{
			linecount = 0;   //Reset linecount on maximum
		} 
		
		switch (linecount)
		{
			case 10: 
				v_2;
				break;
			case 29:
				v_3;
				break;
			case 49: 
				v_4;
				break;
			case 68:
				v_5;
				break;
			case 88: 
				v_6;
				break;
			case 107:
				v_7;
				break;
			case 127: 
				v_8;
				break;
			case 147:
				v_9;
				break;
			case 166: 
				v_10;
				break;
			case 186: 
				v_11;
				break;
			case 205:
				v_12;
				break;
			case 225: 
				v_13;
				break;
			case 244:
				v_14;
				break;
			case 264: 
				v_15;
				break;
			case 283:
				v_16;
				break;
			case 303:
				v_17;
				break;
			case 312:
				v_off;
				break;		
		}
	
	}
	sei();
}


ISR(INT0_vect)
{
// Behandlungsroutine fr INT0 (FBAS)
// Diese Routine wird aufgerufen, wenn ein FBAS-Sync ausgelst wird
// Es wird ein positiver Sync-Impuls passender Lnge erzeugt, der am Ausgang ausgegeben wird

	int i;
	
// Dauer des gesamten Sync-Signals bei FBAS: 12 Mikrosekunden, entspricht 240 Takten 	
// Wert fr Bildlage bercksichtigen
// Es wiurd genaugenommen erst auf das nchste Bild synchronisiert
// Dies ermglicht eine Bildweiteneinstellung ber einen weiten Raum hinaus


	for( i = porch_delay; i < max_delay_time; i++)
		NOP8;

// Sync wird aufgerufen
	hsync_on;
	NOP20;
	NOP20;
	NOP20;
	NOP20;
	NOP20;
//	NOP20;
//	NOP20;
//	NOP20;
//	NOP20;
//	NOP20;
	hsync_off;
}


ISR(INT1_vect)
{
// Behandlungsroutine fr INT1 (VGA, SVGA)
// Diese Routine wird aufgerufen, wenn ein VGA- bzw- SVGA-Sync ausgelst wird
// Es wird ein positiver Sync-Impuls passender Lnge erzeugt, der am Ausgang ausgegeben wird

	int i;

	if (vga_mode == VGA640)
	{
// Dauer des gesamten Sync-Signals bei VGA: 6,7 Mikrosekunden, entspricht 127 Takten 	
// Der Sync-Aufruf erfolgt nach den ersten 13 Takten 
// Wert fr Bildlage bercksichtigen
		for( i = porch_delay; i < max_delay_time; i++)
			NOP8;		

// Sync wird aufgerufen 		
		hsync_on;
		NOP20;

		hsync_off;
	}
	else // vga_mode = VGA800
	{
// Dauer des gesamten Sync-Signals bei SVGA: 6,4 Mikrosekunden, entspricht 124 Takten 	
// Der Sync-Aufruf erfolgt nach den ersten 13 Takten 
// Wert fr Bildlage bercksichtigen
		for( i = porch_delay; i < max_delay_time; i++)
			NOP8;	
// Sync wird aufgerufen 
		hsync_on;
		NOP20;
//		NOP10;
//		NOP4;
		hsync_off;		
	}
	linecount++;
}



ISR(TIMER2_COMPA_vect)
{
	cli();
	INT_CONTROL = INT_DISABLE;
	TIMSK2 = 0;
	if (linecount > 450)  // wird im 76Hz-Takt aufgerufen 450x76Hz=34kHz
		vga_mode = VGA800;
	else	
		vga_mode = VGA640;
	sei();
}	


static void check_vga_mode(void)
{
	linecount = 0;
	vga_mode = 0;
	porch_delay = 31;
	max_delay_time = 30; 
	vga_mode = 0;
	INT_CONTROL = INT1_RISING_EDGE; 
	
	OCR2A = 255; // 19531 = 1 Sekunde bei 20 MHz Takt und Prescaler 1024
	TCCR2B = (1<<CS20) | (1<<CS21) |(1<<CS22) | (1<<WGM22); 	// timer mit prescaler = 1024 starten
	TIMSK2 = (1<<OCIE2A);				// Compare-Interrupt aktivieren

	INT_ENABLE = INT1_ENABLE;
	sei();
	while (vga_mode == 0)
	{
// wird beendet durch Anspringen des Timers	
	}
	if (vga_mode == VGA800)
		PORTD |= (1<<PD0);
	else
		PORTD &= ~(1<<PD0);
	linecount = 0;	
}



static void avr_init(void)
{
	
//Set power mode
  set_sleep_mode(SLEEP_MODE_IDLE); 	
  
//init Ports

// Port C: Output
// Der C-Port enthlt alle notwendigen Ausgangssignale
// C0 - (A) Sync-Signal (Sync positiv)
// C1 - (A) frei
// Die Signale C2 bis C5 dienen als Ausgnge fr Experimente, wenn der interne Generator eingeschaltet ist
// C2 - (A) gleiche Frequenz wie HSync-Signal, aber Takverhltnis 1:1 - nur aktiv, wenn der interne Generator eingeschaltet ist
// C3 - (A) doppelte HSync-Frequenz, nur aktiv, wenn der interne Generator eingeschaltet ist
// C4 - (A) vierfache HSync-Frequenz, nur aktiv, wenn der interne Generator eingeschaltet ist
// C5 - (A) achtfache HSync-Frequenz, nur aktiv, wenn der interne Generator eingeschaltet ist

  DDRC = 0x3F;
  PORTC = 0x00;
  
// Port D: Input
// Der D-Port enthlt alle notwendigen Einstellungssignale
// D0 - (A) Detect-Leitung VGA800, Eingang bei VSync
// D1 - (E) Detect-Leitung VGA640, durch Schalter gewhlt
// D2 - (E) Bild-Sync FBAS, INT0
// D3 - (E) Bild-Sync VGA, INT1
// D4 - (E) frei
// D5 - (E) frei
// D6 - (E) interner Generator aktiv
// D7 - (E) frei

  DDRD = 0x01;		// gem E/A-Angaben
  PORTD = 0xFE;   	// Pull-up-Widerstnde aktivieren und ggf. AUsgnge aktivieren	
  
// Port B: Output bei B1, B2, B3, B4
// Die B-Ports werden fr die LED-Anzeige des gewhlten Betriebsmodus genutzt (VGA800, VGA640, FBAS, intern)
// Sie dienen auerdem zur Zuschaltung von Ladekondensatoren der nachgeschalteten Sgezahnstufen, um bei
// verschiedenen Frequenzen annhernd gleiche Amplituden zu erreichen.
// B0 - (A) vertikaler Sync (bei internem Betrieb)
// B1 - (A) gleiche Frequenz wie VSync-Signal, aber Takverhltnis 1:1 - nur aktiv, wenn der interne Generator eingeschaltet ist
// B2 - (A) doppelte VSync-Frequenz, nur aktiv, wenn der interne Generator eingeschaltet ist
// B3 - (A) vierfache VSync-Frequenz, nur aktiv, wenn der interne Generator eingeschaltet ist (Programmieradapter)
// B4 - (A) achtfache VSync-Frequenz, nur aktiv, wenn der interne Generator eingeschaltet ist (Programmieradapter)
// B5 - (A) (Programmieradapter)
  DDRB = 0x3F;  
  PORTB = 0x00;
  
  return;
}	
	
	
	
	
	


static void Setup_Sync(void)
{
	unsigned int input_values;
	unsigned int detected_sync_mode;

// Daten von Port D lesen unbd auswerten
	input_values = PIND;
	
// Welcher Sync-Mode: Reihenfolge: SVGA, VGA, FBAS, intern	
// nur, wenn kein Signal anliegt, springt der interne Generator an
// Einstellung des Sync-Modes: Schalter D1 unterscheidet zwischen FBAS und VGA
// ist zustzlich Schalter D0 geschlossen, wird auf SVGA optimiert
	if (input_values & VGA640_IN) 
		detected_sync_mode = VGA640;
	else 
		detected_sync_mode = FBAS;
	
// ist der interne Grafikmodus aktiviert? 
// der interne Grafikmodus versetzt das Modul in die Betriebsart "Video-Modul EE3023" bzw. "VGA-Modul"		

	if (input_values & INTERN_GEN)
	{
		if (detected_sync_mode == FBAS)
		{
			detected_sync_mode = INTGEN_FBAS;		// Ausgnge fr VGA640 und intern sind aktiviert
		}
		else
		{
			detected_sync_mode = INTGEN_VGA;
		}	
	}	

// Hat eine Vernderung gegenber dem letzten Setup stattgefunden ?	
	if (detected_sync_mode != sync_mode)
	{
		cli();   // Interrupts abschalten, Signal bricht zusammen
		sync_mode = detected_sync_mode;
		linecount = 0;
		
		PORTB = sync_mode;
		
		INT_CONTROL = INT_DISABLE;  // Interrupts lschen
		TIMSK1 = 0x00;				// Overflow-Timer lschen
		TCCR1B = 0x00;
		
		if (detected_sync_mode == VGA640)
			check_vga_mode();
		
// Interrupt-Funktionen abhngig vom gesetzten Modus aktivieren
// VGA und SVGA syncen ber INT1
// FBAS synct ber INT0
// Interner Sync ber Timer

		switch (sync_mode)
		{
			case VGA640: 
			{
				if (vga_mode == VGA800)
				{
					porch_delay = 3;
					max_delay_time = 37; 
// INT1 wird ausgelst, weitere Verarbeitung im Interrupt-Code		
					INT_CONTROL = INT1_FALLING_EDGE;
				}
				else
				{
					porch_delay = 3;	
					max_delay_time = 46;
// INT1 wird ausgelst, weitere Verarbeitung im Interrupt-Code		
					INT_CONTROL = INT1_RISING_EDGE; 

				}
				PORTB &= ~(1<<PB0); // keine Ausgabe von VSync-Signalen
				INT_ENABLE = INT1_ENABLE;
				break;	
			}	
			
			case FBAS: 
			{
				porch_delay = 1;
				max_delay_time = 99;
// INT0 wird ausgelst, weitere Verarbeitung des Sync-Signals im Interrupt-Code		
				INT_CONTROL = INT0_RISING_EDGE;
				PORTB &= ~(1<<PB0); // keine Ausgabe von VSync-Signalen
				INT_ENABLE = INT0_ENABLE;
				break;
			}
			
			case INTGEN_VGA:
			{
// interner Ablaufzhler wird benutzt und einmal pro Zeile aufgerufen		
// Compare Register
				OCR1A = 636; // Zeilenlnge fr VGA 
				TCCR1B = (1<<CS10) | (1<<WGM12); 	// timer mit prescaler = 1 starten
				TIMSK1 = (1<<OCIE1A);				// Compare-Interrupt aktivieren
				break;
			}
			
			case INTGEN_FBAS:
			{
// interner Ablaufzhler wird benutzt und einmal pro Zeile aufgerufen		
// Compare Register
				OCR1A = 1279; // Zeilenlnge fr FBAS
				TCCR1B = (1<<CS10) | (1<<WGM12); 	// timer mit prescaler = 1 starten
				TIMSK1 = (1<<OCIE1A);				// Compare-Interrupt aktivieren
				break;
			}
			
		}
	}

// enable Interrupts
	sei();	
}		
	
	
	

int main(void)
{
	linecount = 0;
	sync_mode = 0x00;
	vga_mode = 0;
	porch_delay = 0;
	max_delay_time = 25;
	
// Ports initialisieren
	avr_init();
	
    while(1)
	{
		Setup_Sync();
    }
	return(0);
}