// Vertikal-Generator
// fr Philips BRE-Modul 
// Frank Brennecke, 05.12.2013
// Version 1.2
// Fr ATMega88 mit interner 8MHz-Steuerung
// Fuses: Extended F9, High: DE, Low: C2

// horizontale Ausgangsfrequenzen an C0: 37,9 kHz (VGA800), 31,4 kHz (VGA640), 15,625 kHz (FBAS)
// Zustzliche Ausgangsfrequenzen bei internem Modus: fx1, fx2, fx4, fx8 bei Taktverhltnis 1:1

// vertikale Ausgangsfrequenzen an C1: 60 Hz (VGA800), 60 Hz (VGA640), 50 Hz (FBAS)
// Der Vertikalgenerator bentigt aufgrund der geringen internen Frequenzen keinen eigenen Quarz-Oszillator, 
//   sondern kann mit der internen Frequenzeinheit betrieben werden. Beim Betrieb als interner Generator 
//   wird er automatisch mit dem Horizontalgenerator synchronisiert. Er besitzt daher keinen eigenen internen Modus.

#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/io.h>

//vertical Outputs

#define vsync_on PORTC = 0x01; // C0 positiver Sync
#define vsync_off PORTC	= 0x00;

// Modi fr externe Sync-Signale
#define VGA800 		0x1C	// SVGA 800x600 selected - passend fr PORTB/1
#define VGA640 		0x1A 	// VGA 640x480 selected - passend fr PORTB/2
#define FBAS 		0x16 	// FBAS selected - passend fr PORTB/3
// Interner Generator - diese Betriebsart macht aus der Schaltung entweder ein Video-Modul (analog zum EE2023) oder ein VGA-Modul 
#define INTGEN_VGA	0x0A	// Internal VGA Signal Generator active - passend fr PORTB/4
#define INTGEN_FBAS	0x06	// Internal FBAS Signal Generator active - passend fr PORTB/4

#define VGA800_IN	0x01	// SVGA 800x600 erkannt, Eingang D0 (wird vom HSync-Prozessor erkannt)
#define VGA640_IN	0x02 	// VGA 640x480 erkannt, Eingang D1
// FBAS-Signal wird erkannt, wenn D1 "0" ist
#define INTERN_GEN		0x40	// Interner Generator aktiviert, Eingang D6
#define SYNC_INTERN		0x08	// Eingang fr internen Sync, D3

#define SYNC_VGA		0x40	// Eingang fr VGA-Sync, B6
#define SYNC_FBAS		0x80	// Eingang fr FBAS-Sync, B7

#define VGA640_lines_internal  	525 // VGA-Signal hat 525 Zeilen
#define FBAS_lines_internal	 	313 // FBAS-Signal hat 2x312,5 Zeilen

#define INT_CONTROL			EICRA
#define INT0_FALLING_EDGE	0x02 	//Bit 1 gesetzt 
#define INT0_RISING_EDGE	0x03	//Bit 0 und Bit 1 gesetzt
#define INT1_FALLING_EDGE	0x08	//Bit 3 gesetzt
#define INT1_RISING_EDGE	0x0C	//Bit 2 und Bit 3 gesetzt

#define INT_ENABLE			EIMSK
#define INT0_ENABLE			0x01	//Bit 0 gesetzt
#define INT1_ENABLE			0x02	//Bit 1 gesetzt
#define INT_DISABLE			0x00	//kein Bit gesetzt


//Global variables
volatile unsigned int sync_mode;   // Welcher Sync-Modus ist eingestellt?
volatile unsigned int max_delay_time; // Maximale Verschiebung der Sync-Position (Bildlage)
volatile unsigned int porch_delay;  // Tatschliche Verschiebung der Sync-Position (Bildlage)
volatile unsigned int sync_avail; // Externer Sync verfgbar? Prfung fr Umschaltung auf interne Synchronisation


   
ISR(INT0_vect)
{
// Dieser Interrupt wird fr jedes Halbbild des FBAS-Signals einmal ausgelst (50 Hz = 20ms)
	int i;
	
	cli();
	sync_avail++;		// Sync-Signal liegt an
	
	// Wert fr Bildlage bercksichtigen
	for( i = porch_delay; i < max_delay_time; i++)
		_delay_ms(5);
	
// vertical Sync	
	vsync_on;
    _delay_ms(10);		// FBAS-Timing
	vsync_off;	
	sei();
}   
   

ISR(INT1_vect)
{
// Dieser Interrupt wird fr jedes Bild des VGA-Signals einmal ausgelst (60 bzw. 56 Hz)
	int i;
	cli();				// Interrupts abschalten
	sync_avail++;		// Sync-Signal liegt an

	if (sync_mode == VGA800)
	{
// Wert fr Bildlage bercksichtigen	
		for( i = porch_delay; i < max_delay_time; i++)
			_delay_ms(2);
		vsync_on;
		_delay_ms(2);	// SVGA-Timing
		vsync_off;	
	}
	else
	{
// Wert fr Bildlage bercksichtigen	
	    for( i = porch_delay; i < max_delay_time; i++)
			_delay_ms(2);
		vsync_on;
		_delay_ms(2);	// VGA-Timing
		vsync_off;
	}	
	sei();  // Interrupts wieder aktivieren
}



ISR(TIMER1_COMPA_vect)
{
	cli();
// Behandlungsroutine fr Ablaufzhler, um Verfgbarkeit von Sync-Signalen zu prfen
	if (sync_avail > 3)
	{
		PORTD &= ~(1<<PD6);	// Interner Generator deaktiviert
		PORTB |= (1<<PB4);		// LED deaktivieren (Achtung, LED arbeitet invertiert)
	}	
	else
	{
		PORTD |= (1<<PD6);		// Interner Generator eingeschaltet
		PORTB &= ~(1<<PB4);	// LED aktivieren (Achtung, LED arbeitet invertiert)
	}
	sync_avail = 0;	
	sei();
}



static void avr_init(void)
{
	
//Set power mode
  set_sleep_mode(SLEEP_MODE_IDLE); 	
  
//init Ports

// Port C: Output
// Der C-Port enthlt alle notwendigen Ausgangssignale
// C0 - (A) Sync-Signal (Sync positiv)
// C1 - (A) frei
// C2 - (A) frei
// C3 - (A) frei
// C4 - (A) frei
// C5 - (A) frei
  DDRC = 0x3F;
  PORTC = 0x00;
  
// Port D: Input
// Der D-Port enthlt alle notwendigen Einstellungssignale
// D0 - (E) Detect-Leitung VGA800, Eingang bei VSync
// D1 - (E) Detect-Leitung VGA640, durch Schalter gewhlt
// D2 - (E) Bild-Sync FBAS, INT0
// D3 - (E) Bild-Sync VGA, INT1
// D4 - (E) frei
// D5 - (E) frei
// D6 - (A) interner Generator aktiv
// D7 - (E) frei

  DDRD = 0x40;		// gem E/A-Angaben
  PORTD = 0xFF;   	// Pull-up-Widerstnde aktivieren und ggf. AUsgnge aktivieren	
  
// Port B: Output bei B1, B2, B3, B4
// Die B-Ports werden fr die LED-Anzeige des gewhlten Betriebsmodus genutzt (VGA800, VGA640, FBAS, intern)
// Sie dienen auerdem zur Zuschaltung von Ladekondensatoren der nachgeschalteten Sgezahnstufen, um bei
// verschiedenen Frequenzen annhernd gleiche Amplituden zu erreichen.
// B0 - (A) frei
// B1 - (A) VGA800
// B2 - (A) VGA640
// B3 - (A) FBAS (Programmieradapter)
// B4 - (A) interner Generator aktiv - immer zusammen mit B2 oder B3 aktiv (Programmieradapter)
// B5 - (A) (Programmieradapter)
  DDRB = 0x3F;  
  PORTB = 0x00;
  
  return;
}



static void Setup_Sync(void)
{
	unsigned int input_values;
	unsigned int detected_sync_mode;

// Daten von Port D lesen und auswerten
	input_values = PIND;
	
// Einstellung des Sync-Modes: Schalter D1 unterscheidet zwischen FBAS und VGA
// Keine automatische Erkennung, damit auch intern zwischen FBAS und VGA gewhlt werden kann
// ist zustzlich D0 gesetzt, wird auf SVGA optimiert
	if (input_values & VGA640_IN) 
	{
		if (input_values & VGA800_IN) 
			detected_sync_mode = VGA800;
		else
			detected_sync_mode = VGA640;
	}
	else 
		detected_sync_mode = FBAS;
		
		
		
// Hat eine Vernderung gegenber dem letzten Setup stattgefunden ?	
	if (detected_sync_mode != sync_mode)
	{
		cli();   // Interrupts abschalten, Signal bricht zusammen
		INT_ENABLE = 0;
		sync_mode = detected_sync_mode;
		
		switch (sync_mode)
		{
			case VGA800: 
			{
				PORTB |= (1<<PB2); 		// Achtung, LEDs arbeiten invertiert
				PORTB |= (1<<PB3); 
				PORTB &= ~(1<<PB1);
				max_delay_time = 25;
				porch_delay = max_delay_time + 1; 
				INT_CONTROL = INT1_RISING_EDGE; // INT0 rising edge, positive pulse 
				INT_ENABLE = INT1_ENABLE;
				break;
			}
			
			case VGA640: 
			{
				PORTB |= (1<<PB1);			// Achtung, LEDs arbeiten invertiert
				PORTB |= (1<<PB3);
				PORTB &= ~(1<<PB2);
				max_delay_time = 64;
				porch_delay = 61;
				INT_CONTROL = INT1_RISING_EDGE; // INT0 rising edge, negative pulse 
				INT_ENABLE = INT1_ENABLE;	
				break;
			}
		
			case FBAS: 
			{
				PORTB |= (1<<PB1);			// Achtung, LEDs arbeiten invertiert
				PORTB |= (1<<PB2);
				PORTB &= ~(1<<PB3);
				max_delay_time = 30;
				porch_delay = 29;
				INT_CONTROL = INT0_RISING_EDGE;
				INT_ENABLE = INT0_ENABLE;
				break;
			}
			
			default:
				break;
			
		}
		
		// Ablauftimer prft, ob Sync-Signale anliegen		
		OCR1A = 2000; // 19531 = 1 Sekunde bei 20 MHz Takt und Prescaler 1024
		TCCR1B = (1<<CS10) | (1<<CS12) | (1<<WGM12); 	// timer mit prescaler = 1024 starten
		TIMSK1 = (1<<OCIE1A);				// Compare-Interrupt aktivieren
	}

// enable Interrupts
	sei();	
	
}		





int main(void)
{
	sync_mode = 0x00;
	sync_avail = 0;
// Ports initialisieren
	avr_init();
    while(1)
	{
		Setup_Sync();
    }
	return(0);
}





	
    